/*
	The original file "etherpci.c" from Be Sample Code.

	----------------------
	Be Sample Code License
	----------------------

	Copyright 1991-1999, Be Incorporated.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions, and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright
	   notice, this list of conditions, and the following disclaimer in the
	   documentation and/or other materials provided with the distribution.

	3. The name of the author may not be used to endorse or promote products
	   derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
	IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
	OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
	PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
	AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
	OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    
*/

/* Others from Linux device driver for the mtd80x Ethernet chip. */

/*
	Written 1998-2001 by Donald Becker <becker at scyld.com>.

	This software may be used and distributed according to the terms of
	the GNU General Public License (GPL), incorporated herein by reference.
	Drivers based on or derived from this code fall under the GPL and must
	retain the authorship, copyright and license notice.  This file is not
	a complete program and may only be used when the entire operating
	system is licensed under the GPL.

	The author may be reached as becker@scyld.com, or C/O
	Scyld Computing Corporation
	410 Severn Ave., Suite 210
	Annapolis MD 21403
*/

/*
	2006.11.01		Modified for BeOS by Anthony Lee <don.anthony.lee at gmail.com>.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <bsd_mem.h>

#include <SupportDefs.h>
#include <OS.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <PCI.h>
#include <ByteOrder.h>

#define ENABLE_DEBUG		0

#define kDevName			"mtd80x"
#define kDevDir				"net/" kDevName "/"
#define DEVNAME_LENGTH		28
#define MAX_CARDS			8
#define MAX_MULTI			32
#define MAX_QUEUE_LENGTH	1024

#if ENABLE_DEBUG
#ifdef __GNUC__
#define DEBUG(x...)			dprintf(kDevName ": " __FUNCTION__ " --- " x)
#else
#define DEBUG(x...)			dprintf(kDevName ": " x)
#endif
#else
#ifdef __GNUC__
#define DEBUG(x...)			(void)0
#else
inline void DEBUG(const char *format, ...)
{
}
#endif
#endif /* ENABLE_DEBUG */

static pci_module_info *gPCIModInfo = NULL;
static char *gDevNameList[MAX_CARDS + 1];
static pci_info *gDevList[MAX_CARDS + 1];
static int32 gOpenMask = 0;

static int32 options[MAX_CARDS] = {-1, -1, -1, -1, -1, -1, -1, -1};
static int32 full_duplex[MAX_CARDS] = {-1, -1, -1, -1, -1, -1, -1, -1};
static uint32 mtu = 0;
static int32 max_interrupt_work = 20;

#define TX_RING_SIZE	64
#define RX_RING_SIZE	128
#define PKT_BUF_SZ	1520

/* Wait this long before giving up on an ethernet transmit */
#define ETHER_TRANSMIT_TIMEOUT		((bigtime_t)1000000)
#define TX_TIMEOUT					((bigtime_t)10000)

/* Chips prior to the 80x have an external MII transceiver. */
enum chip_capability_flags
{
	HasMIIXcvr = 1,
	HasChipXcvr = 2,
};

typedef struct pci_id_info
{
	char name[256];
	uint32 id;
	uint32 flags;
} pci_id_info;

static pci_id_info supported_devs[] =
{
	{"SURECOM EP-320X-S 100/10M Ethernet PCI Adapter", 0x08031516, HasChipXcvr},
};

static int32 n_supported_devs = sizeof(supported_devs) / sizeof(pci_id_info);

/*
 * ioctls: belongs in a public header file
 * somewhere, so that the net_server and other ethernet drivers can use.
 */
enum
{
	ETHER_GETADDR = B_DEVICE_OP_CODES_END,	/* get ethernet address */
	ETHER_INIT,								/* set irq and port */
	ETHER_NONBLOCK,							/* set/unset nonblocking mode */
	ETHER_ADDMULTI,							/* add multicast addr */
	ETHER_REMMULTI,							/* rem multicast addr */
	ETHER_SETPROMISC,						/* set promiscuous */
	ETHER_GETFRAMESIZE,						/* get frame size */
};

typedef struct mac_address
{
	unsigned char byte[6];
} mac_address;

/* The Rx and Tx buffer descriptors. */
/* Note that using only 32 bit fields simplifies conversion to big-endian architectures. */
typedef struct netdev_desc_tx
{
	uint32 status;
	uint32 ctrl_length;
	uint32 buf_addr;
	uint32 next_desc;
} netdev_desc_tx;

typedef struct netdev_desc_rx
{
	uint32 status;
	uint32 ctrl_length;
	uint32 buf_addr;
	uint32 next_desc;
	unsigned char *rxbuffer;
	uint32 reserved1;
	uint32 reserved2;
	uint32 reserved3;
} netdev_desc_rx;

typedef struct _netdev_list netdev_list;
struct _netdev_list
{
	uint32 size;
	unsigned char *buf;
	netdev_list *prev;
	netdev_list *next;
};

typedef struct mtd80x_private
{
	int32		devID;				/* device identifier: 0-n */
	pci_info	*pciInfo;
	pci_id_info	*pciId;

	uint32		reg_base;			/* hardware register base address */
#if !__INTEL__
	area_id		ioarea;				/* Area used for MMIO of hostRegs */
#endif

	thread_id	iothr;
	sem_id		iosem;

	sem_id		iolock;				/* ethercard io */

	bool		interrupted;		/* interrupted system call */
	int32		inrw;				/* in read or write function */
	int			nonblocking;		/* non-blocking mode */
	int			promisc;
	int			all_multi;

	sem_id		ilock;				/* waiting for input */
	sem_id		olock;				/* waiting to output */

	mac_address	dev_addr;
	uint32		nmulti;				/* number of multicast addresses */
	mac_address	multi[MAX_MULTI];	/* multicast addresses */

	uint8		full_duplex;		/* Full-duplex operation requested. */
	uint8		duplex_lock;
	uint8		medialock;			/* Do not sense media. */
	uint32		default_port;		/* Last dev->if_port value. */
	uint32		mtu;

	uint32		mii_cnt;			/* MII device addresses. */
	uint16		advertising;		/* NWay media advertisement */
	uint32		phys[2];			/* MII device addresses. */

	area_id		rx_area;
	area_id		tx_area;

	netdev_desc_rx	*rx_ring;
	netdev_desc_tx	*tx_ring;

	uint32		rx_ring_addr;
	uint32		tx_ring_addr;

	uint32		rx_buf_sz;			/* Based on MTU+slack. */
	uint32		mcast_filter[2];
	uint32		txrx_config;

	netdev_list	*rx_list;
	uint32		rx_list_len;

	uint32		txfreecount;
	bigtime_t	last_trans;
} mtd80x_private;

#if __INTEL__
#define readb(addr)					(*gPCIModInfo->read_io_8)(addr)
#define readw(addr)					(*gPCIModInfo->read_io_16)(addr)
#define readl(addr)					(*gPCIModInfo->read_io_32)(addr)
#define writeb(value, addr)			(*gPCIModInfo->write_io_8)(addr, (value))
#define writew(value, addr)			(*gPCIModInfo->write_io_16)(addr, (value))
#define writel(value, addr)			(*gPCIModInfo->write_io_32)(addr, (value))
#define cpu_to_le32(value)			(value)
#define le32_to_cpu(value)			(value)
#else /* PPC */
#define readb(addr)					(__eieio(), (*((volatile uint8*)(addr))))
#define readw(addr)					(__eieio(), B_LENDIAN_TO_HOST_INT16((*((volatile uint16*)(addr)))))
#define readl(addr)					(__eieio(), B_LENDIAN_TO_HOST_INT32((*((volatile uint32*)(addr)))))
#define writeb(value, addr)			(__eieio(), (*((volatile uint8 *)(addr)) = (value)))
#define writew(value, addr)			(__eieio(), (*((volatile uint16*)(addr)) = B_HOST_TO_LENDIAN_INT16(value)))
#define writel(value, addr)			(__eieio(), (*((volatile uint32*)(addr)) = B_HOST_TO_LENDIAN_INT32(value)))
#define cpu_to_le32(value)			B_HOST_TO_LENDIAN_INT32(value)
#define le32_to_cpu(value)			B_LENDIAN_TO_HOST_INT32(value)
#endif

/* Offsets to the various registers.
   Most accesses must be longword aligned. */
enum register_offsets {
	StationAddr			= 0x00,
	MulticastFilter0	= 0x08,
	MulticastFilter1	= 0x0C,
	FlowCtrlAddr		= 0x10,
	RxConfig			= 0x18,
	TxConfig			= 0x1a,
	PCIBusCfg			= 0x1c,
	TxStartDemand		= 0x20,
	RxStartDemand		= 0x24,
	RxCurrentPtr		= 0x28,
	TxRingPtr			= 0x2c,
	RxRingPtr			= 0x30,
	IntrStatus			= 0x34,
	IntrEnable			= 0x38,
	FlowCtrlThreshold	= 0x3c,
	MIICtrl				= 0x40,
	EECtrl				= 0x40,
	RxErrCnts			= 0x44,
	TxErrCnts			= 0x48,
	PHYMgmt				= 0x4c,
};

/* Bits in the interrupt status/mask registers. */
enum intr_status_bits {
	IntrRxErr		= 0x0002,
	IntrRxDone		= 0x0004,
	IntrTxDone		= 0x0008,
	IntrTxEmpty		= 0x0010,
	IntrRxEmpty		= 0x0020,
	StatsMax		= 0x0040,
	RxEarly			= 0x0080,
	TxEarly			= 0x0100,
	RxOverflow		= 0x0200,
	TxUnderrun		= 0x0400,
	IntrPCIErr		= 0x2000,
	NWayDone		= 0x4000,
	LinkChange		= 0x8000,
};

/* Bits in the RxMode (np->txrx_config) register. */
enum rx_mode_bits {
	RxEnable		= 0x01,
	RxFilter		= 0xfe,
	AcceptErr		= 0x02,
	AcceptRunt		= 0x08,
	AcceptBroadcast	= 0x40,
	AcceptMulticast	= 0x20,
	AcceptAllPhys	= 0x80,
	AcceptMyPhys	= 0x00,
	RxFlowCtrl		= 0x2000,
	TxEnable		= 0x40000,
	TxModeFDX		= 0x00100000,
	TxThreshold		= 0x00e00000,
};

/* Misc. bits. */
enum misc_bits {
	BCR_Reset		= 1,		/* PCIBusCfg */
};

/* Bits in network_desc.status */
enum desc_status_bits {
	DescOwn				= 0x80000000,
	RxDescStartPacket	= 0x0800,
	RxDescEndPacket		= 0x0400,
	RxDescWholePkt		= 0x0c00,
	RxDescErrSum		= 0x80,
	RxErrRunt			= 0x40,
	RxErrLong			= 0x20,
	RxErrFrame			= 0x10,
	RxErrCRC			= 0x08,
	RxErrCode			= 0x04,
	TxErrAbort			= 0x2000,
	TxErrCarrier		= 0x1000,
	TxErrLate			= 0x0800,
	TxErr16Colls		= 0x0400,
	TxErrDefer			= 0x0200,
	TxErrHeartbeat		= 0x0100,
	TxColls				= 0x00ff,
};

/* Bits in network_desc.ctrl_length */
enum ctrl_length_bits {
	TxIntrOnDone		= 0x80000000,
	TxIntrOnFIFO		= 0x40000000,
	TxDescEndPacket		= 0x20000000,
	TxDescStartPacket	= 0x10000000,
	TxAppendCRC			= 0x08000000,
	TxPadTo64			= 0x04000000,
	TxNormalPkt			= 0x3C000000,
};


/* Read the EEPROM and MII Management Data I/O (MDIO) interfaces.  These are
   often serial bit streams generated by the host processor.
   The example below is for the common 93c46 EEPROM, 64 16 bit words. */

/* This "delay" forces out buffered PCI writes. */
#define eeprom_delay(ee_addr)			readl(ee_addr); snooze(2000); readl(ee_addr)

enum EEPROM_Ctrl_Bits {
	EE_ShiftClk		= 0x04 << 16,
	EE_ChipSelect	= 0x88 << 16,
	EE_DataOut		= 0x02 << 16,
	EE_DataIn		= 0x01 << 16,
	EE_Write0		= 0x88 << 16,
	EE_Write1		= 0x8a << 16,
};

/* The EEPROM commands always start with 01.. preamble bits.
   Commands are prepended to the variable-length address. */
enum EEPROM_Cmds
{
	EE_WriteCmd	= 5,
	EE_ReadCmd	= 6,
	EE_EraseCmd	= 7,
};


static int eeprom_read(uint32 addr, int location)
{
	int i;
	int retval = 0;
	uint32 ee_addr = addr + EECtrl;
	int read_cmd = location | (EE_ReadCmd << 6);

	writel(EE_ChipSelect, ee_addr);

	/* Shift the read command bits out. */
	for(i = 10; i >= 0; i--)
	{
		int dataval = (read_cmd & (1 << i)) ? EE_Write1 : EE_Write0;
		writel(dataval, ee_addr);
		eeprom_delay(ee_addr);
		writel(dataval | EE_ShiftClk, ee_addr);
		eeprom_delay(ee_addr);
	}
	writel(EE_ChipSelect, ee_addr);
	eeprom_delay(ee_addr);

	for(i = 16; i > 0; i--)
	{
		writel(EE_ChipSelect | EE_ShiftClk, ee_addr);
		eeprom_delay(ee_addr);
		retval = (retval << 1) | ((readl(ee_addr) & EE_DataIn) ? 1 : 0);
		writel(EE_ChipSelect, ee_addr);
		eeprom_delay(ee_addr);
	}

	/* Terminate the EEPROM access. */
	writel(EE_ChipSelect, ee_addr);
	writel(0, ee_addr);
	return retval;
}


/*  MII transceiver control section.
    Read and write the MII registers using software-generated serial
    MDIO protocol.  See the MII specifications or DP83840A data sheet
    for details.

    The maximum data clock rate is 2.5 Mhz.  The minimum timing is usually
    met by back-to-back 33Mhz PCI cycles. */
#define mdio_delay(mdio_addr) readl(mdio_addr)

/* Set iff a MII transceiver on any interface requires mdio preamble.
   This only set with older tranceivers, so the extra
   code size of a per-interface flag is not worthwhile. */
static char mii_preamble_required = 0;

enum mii_reg_bits {
	MDIO_ShiftClk	= 0x0001,
	MDIO_Data		= 0x0002,
	MDIO_EnbOutput	= 0x0004,
};

#define MDIO_EnbIn  (0)
#define MDIO_WRITE0 (MDIO_EnbOutput)
#define MDIO_WRITE1 (MDIO_Data | MDIO_EnbOutput)


/* Generate the preamble required for initial synchronization and
   a few older transceivers. */
static void mdio_sync(uint32 mdio_addr)
{
	int bits = 32;

	/* Establish sync by sending at least 32 logic ones. */
	while(--bits >= 0)
	{
		writel(MDIO_WRITE1, mdio_addr);
		mdio_delay(mdio_addr);
		writel(MDIO_WRITE1 | MDIO_ShiftClk, mdio_addr);
		mdio_delay(mdio_addr);
	}
}


static int mdio_read(mtd80x_private *np, uint32 phy_id, uint32 location)
{
	uint32 ioaddr = np->reg_base;
	uint32 mdio_addr = ioaddr + MIICtrl;
	int mii_cmd = (0xf6 << 10) | (phy_id << 5) | location;
	int i, retval = 0;

	if(location >= 32) return 0xffff;

	if(phy_id >= 32)
	{
		if(location < 6) return readw(ioaddr + PHYMgmt + location * 2);
		else if (location == 16) return readw(ioaddr + PHYMgmt + 6 * 2);
		else if (location == 17) return readw(ioaddr + PHYMgmt + 7 * 2);
		else if (location == 18) return readw(ioaddr + PHYMgmt + 10 * 2);
		else return 0;
	}

	if(mii_preamble_required) mdio_sync(mdio_addr);

	/* Shift the read command bits out. */
	for(i = 15; i >= 0; i--)
	{
		int dataval = (mii_cmd & (1 << i)) ? MDIO_WRITE1 : MDIO_WRITE0;

		writel(dataval, mdio_addr);
		mdio_delay(mdio_addr);
		writel(dataval | MDIO_ShiftClk, mdio_addr);
		mdio_delay(mdio_addr);
	}

	/* Read the two transition, 16 data, and wire-idle bits. */
	for(i = 19; i > 0; i--)
	{
		writel(MDIO_EnbIn, mdio_addr);
		mdio_delay(mdio_addr);
		retval = (retval << 1) | ((readl(mdio_addr) & MDIO_Data) ? 1 : 0);
		writel(MDIO_EnbIn | MDIO_ShiftClk, mdio_addr);
		mdio_delay(mdio_addr);
	}

	return((retval>>1) & 0xffff);
}


static void mdio_write(mtd80x_private *np, uint32 phy_id, uint32 location, int value)
{
	uint32 ioaddr = np->reg_base;
	uint32 mdio_addr = ioaddr + MIICtrl;
	int mii_cmd = (0x5002 << 16) | (phy_id << 23) | (location << 18) | value;
	int i;

	if(location == 4  && phy_id == np->phys[0]) np->advertising = value;
	else if(location >= 32) return;

	if(phy_id == 32)
	{
		if(location < 6) writew(value, ioaddr + PHYMgmt + location * 2);
		else if (location == 16) writew(value, ioaddr + PHYMgmt + 6 * 2);
		else if (location == 17) writew(value, ioaddr + PHYMgmt + 7 * 2);
		return;
	}

	if(mii_preamble_required) mdio_sync(mdio_addr);

	/* Shift the command bits out. */
	for(i = 31; i >= 0; i--)
	{
		int dataval = (mii_cmd & (1 << i)) ? MDIO_WRITE1 : MDIO_WRITE0;

		writel(dataval, mdio_addr);
		mdio_delay(mdio_addr);
		writel(dataval | MDIO_ShiftClk, mdio_addr);
		mdio_delay(mdio_addr);
	}

	/* Clear out extra bits. */
	for(i = 2; i > 0; i--)
	{
		writel(MDIO_EnbIn, mdio_addr);
		mdio_delay(mdio_addr);
		writel(MDIO_EnbIn | MDIO_ShiftClk, mdio_addr);
		mdio_delay(mdio_addr);
	}

	return;
}


static int32 get_pci_list(pci_info **info, int32 maxEntries)
{
	int32 i, j, entries = 0;
	pci_info item;

	if(gPCIModInfo == NULL || info == NULL) return 0;

	for(i = 0; entries < maxEntries; i++)
	{
		bzero(&item, sizeof(pci_info));

		if(gPCIModInfo->get_nth_pci_info(i, &item) != B_OK) break;

		for(j = 0; j < n_supported_devs; j++)
		{
			if((supported_devs[j].id & 0xffff) == item.vendor_id &&
			   ((supported_devs[j].id >> 16) & 0xffff) == item.device_id) break;
		}
		if(j == n_supported_devs) continue;

		/* check if the device really has an IRQ */
		if((item.u.h0.interrupt_line == 0) || (item.u.h0.interrupt_line == 0xff))
		{
			DEBUG("found with invalid IRQ - check IRQ assignement\n");
			continue;
		}

		if((*info = (pci_info*)malloc(sizeof(pci_info))) == NULL)
		{
			DEBUG("memory allocate failed!\n");
			continue;
		}

		DEBUG("found %s at IRQ %x\n", supported_devs[j].name, item.u.h0.interrupt_line);
		memcpy(*info++, &item, sizeof(pci_info));
		entries++;
	}

	*info = NULL;

	return entries;
}


static status_t free_pci_list(pci_info **info)
{
	if(info == NULL) return B_ERROR;
	while(*info != NULL) free(*info++);
	return B_OK;
}


static status_t enable_addressing(mtd80x_private *data)
{
	unsigned char cmd;

#if __INTEL__
	data->reg_base = data->pciInfo->u.h0.base_registers[0];
#else /* PPC */
	int32 base, size, offset;

	base = data->pciInfo->u.h0.base_registers[0];
	size = data->pciInfo->u.h0.base_register_sizes[0];

	/* Round down to nearest page boundary */
	base = base & ~(B_PAGE_SIZE-1);

	/* Adjust the size */
	offset = data->pciInfo->u.h0.base_registers[0] - base;
	size += offset;
	size = (size + (B_PAGE_SIZE - 1)) & ~(B_PAGE_SIZE - 1);

	DEBUG("PCI base = %x size = %x offset = %x\n", base, size, offset);

	if((data->ioarea = map_physical_memory(kDevName "_regs", (void *)base, size,
					       B_ANY_KERNEL_ADDRESS, B_READ_AREA | B_WRITE_AREA,
					       (void **)&data->reg_base)) < 0) return B_ERROR;

	data->reg_base = data->reg_base + offset;
#endif

	DEBUG("reg_base = %x\n", data->reg_base);

	/* enable pci address & memory access */
	cmd = (gPCIModInfo->read_pci_config)(data->pciInfo->bus, data->pciInfo->device, data->pciInfo->function, PCI_command, 2);
	cmd |= PCI_command_io | PCI_command_memory;
	(gPCIModInfo->write_pci_config)(data->pciInfo->bus, data->pciInfo->device, data->pciInfo->function, PCI_command, 2, cmd);

	return B_OK;
}


static int probe(mtd80x_private *np)
{
	uint32 ioaddr = np->reg_base;
	int i, option = options[np->devID];

	for(i = 0; i < n_supported_devs; i++)
		if(supported_devs[i].id == ((uint32)np->pciInfo->vendor_id >> 16) | ((uint32)np->pciInfo->device_id << 16)) break;
	if(i == n_supported_devs) return 0;
	np->pciId = &(supported_devs[i]);

	/* Get MAC Address */
	for(i = 0; i < 3; i++) ((uint16*)&(np->dev_addr))[i] = eeprom_read(ioaddr, i + 8);
	if(memcmp(&(np->dev_addr), "\0\0\0\0\0\0", 6) == 0)
	{
		DEBUG("MISSING EEPROM ADDRESS.\n");
		return 0;
	}
	else
	{
		DEBUG("MAC address - %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n",
		      np->dev_addr.byte[0], np->dev_addr.byte[1], np->dev_addr.byte[2],
		      np->dev_addr.byte[3], np->dev_addr.byte[4], np->dev_addr.byte[5]);
	}

	/* Reset the chip to erase previous misconfiguration. */
	writel(BCR_Reset, ioaddr + PCIBusCfg);

	/* The lower four bits are the media type. */
	if(option > 0)
	{
		if(option & 0x220) np->full_duplex = 1;
		np->default_port = option & 0x3ff;
		if(np->default_port) np->medialock = 1;
	}

	if(full_duplex[np->devID] > 0) np->full_duplex = 1;

	if(np->full_duplex)
	{
		DEBUG("Set to forced full duplex, autonegotiation disabled.\n");
		np->duplex_lock = 1;
	}

	if(mtu > 0) np->mtu = mtu;

	if(np->pciId->flags & HasMIIXcvr)
	{
		int phy, phy_idx = 0;
		for (phy = 0; phy < 32 && phy_idx < 4; phy++)
		{
			int mii_status = mdio_read(np, phy, 1);
			if(mii_status != 0xffff  &&  mii_status != 0x0000)
			{
				np->phys[phy_idx++] = phy;
				np->advertising = mdio_read(np, phy, 4);
				DEBUG("MII PHY found at address %d, status 0x%4.4x advertising %4.4x.\n",
				      phy, mii_status, np->advertising);
			}
		}
		np->mii_cnt = phy_idx;
	}
	if(np->pciId->flags & HasChipXcvr)
	{
		np->phys[np->mii_cnt++] = 32;
		DEBUG("Internal PHY status 0x%4.4x advertising %4.4x.\n",
		      mdio_read(np, 32, 1), mdio_read(np, 32, 4));
	}

	if(np->default_port & 0x330)
	{
		np->medialock = 1;
		DEBUG("Forcing %dMbs %s-duplex operation.\n",
		      (option & 0x300 ? 100 : 10), (option & 0x220 ? "full" : "half"));
		if(np->mii_cnt)
			mdio_write(np, np->phys[0], 0,
				       ((option & 0x300) ? 0x2000 : 0) | /* 100mbps? */
				       ((option & 0x220) ? 0x0100 : 0)); /* Full duplex? */
	}

	return 1;
}


static uint32 get_phy_address(void *addr, size_t size)
{
	physical_entry entry;

	if(addr == NULL || size == 0) return 0;
	if(get_memory_map(addr, size, &entry, 1) != B_OK) return 0;

	return (uint32)entry.address;
}


static int32 mtd80x_thread(void *_data);
static status_t allocate_resources(mtd80x_private *data)
{
	uint32 buf_size;

	if((data->iothr = spawn_kernel_thread(mtd80x_thread, NULL, B_NORMAL_PRIORITY, data)) < 0) return B_ERROR;

	data->iosem = create_sem(0, "mtd80x intr");
	set_sem_owner(data->iosem, B_SYSTEM_TEAM);
	data->iolock = create_sem(1, "mtd80x io");
	set_sem_owner(data->iolock, B_SYSTEM_TEAM);
	data->txfreecount = TX_RING_SIZE;
	data->olock = create_sem(data->txfreecount, "mtd80x output");
	set_sem_owner(data->olock, B_SYSTEM_TEAM);
	data->ilock = create_sem(0, "mtd80x input");
	set_sem_owner(data->ilock, B_SYSTEM_TEAM);

	data->interrupted = false;
	data->inrw = 0;
	data->nonblocking = 0;
	data->promisc = 0;
	data->all_multi = 0;

	data->rx_list = NULL;
	data->rx_list_len = 0;

	data->rx_buf_sz = (data->mtu <= 1500 ? PKT_BUF_SZ : data->mtu + 32);

	buf_size = (data->rx_buf_sz + sizeof(netdev_desc_rx)) * RX_RING_SIZE;
	buf_size = (buf_size + (B_PAGE_SIZE - 1)) & ~(B_PAGE_SIZE - 1);
	data->rx_area = create_area("mtd80x rx buffer", (void**)&(data->rx_ring), B_ANY_KERNEL_ADDRESS,
				    buf_size, B_FULL_LOCK | B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
	data->rx_ring_addr = get_phy_address(data->rx_ring, buf_size);
	bzero(data->rx_ring, buf_size);

	buf_size = (PKT_BUF_SZ + sizeof(netdev_desc_tx)) * TX_RING_SIZE;
	buf_size = (buf_size + (B_PAGE_SIZE - 1)) & ~(B_PAGE_SIZE - 1);
	data->tx_area = create_area("mtd80x tx buffer", (void**)&(data->tx_ring), B_ANY_KERNEL_ADDRESS,
				    buf_size, B_FULL_LOCK | B_CONTIGUOUS, B_READ_AREA | B_WRITE_AREA);
	data->tx_ring_addr = get_phy_address(data->tx_ring, buf_size);
	bzero(data->tx_ring, buf_size);

	return B_OK;
}


static void free_resources(mtd80x_private *data)
{
	netdev_list *list, *tmp;

	delete_sem(data->iosem);
	delete_sem(data->iolock);
	delete_sem(data->olock);
	delete_sem(data->ilock);
	delete_area(data->rx_area);
	delete_area(data->tx_area);

	for(list = data->rx_list; list != NULL;) {tmp = list; list = list->next; if(tmp->buf) free(tmp->buf); free(tmp);}
}


/* Initialize the Rx and Tx rings */
static void init_ring(mtd80x_private *np)
{
	int i;

	/* Initialize all Rx descriptors. */
	for(i = 0; i < RX_RING_SIZE; i++)
	{
		np->rx_ring[i].ctrl_length = cpu_to_le32(np->rx_buf_sz);
		np->rx_ring[i].status = cpu_to_le32(DescOwn);
		np->rx_ring[i].next_desc = cpu_to_le32(np->rx_ring_addr + (i == RX_RING_SIZE - 1 ? 0 : ((i + 1) * sizeof(netdev_desc_rx))));
		np->rx_ring[i].rxbuffer = ((unsigned char*)(np->rx_ring) + RX_RING_SIZE * sizeof(netdev_desc_rx) + np->rx_buf_sz * i);
		np->rx_ring[i].buf_addr = cpu_to_le32(np->rx_ring_addr + RX_RING_SIZE * sizeof(netdev_desc_rx) + np->rx_buf_sz * i);
	}

	for(i = 0; i < TX_RING_SIZE; i++)
	{
		np->tx_ring[i].ctrl_length = 0;
		np->tx_ring[i].status = 0;
		np->tx_ring[i].next_desc = cpu_to_le32(np->tx_ring_addr + (i == TX_RING_SIZE - 1 ? 0 : ((i + 1) * sizeof(netdev_desc_tx))));
		np->tx_ring[i].buf_addr = 0;
	}

	return;
}


static uint32 ether_crc(int length, unsigned char *data)
{
	uint32 crc = ~0;

	while(--length >= 0)
	{
		unsigned char current_octet = *data++;
		int bit;
		for(bit = 0; bit < 8; bit++, current_octet >>= 1)
			crc = (crc << 1) ^ ((crc < 0) ^ (current_octet & 1) ? 0x04c11db7U : 0);
	}

	return crc;
}


static void set_rx_mode(mtd80x_private *np)
{
	uint32 ioaddr = np->reg_base;
	uint32 mc_filter[2]; /* Multicast hash filter */
	uint32 rx_mode;
	int i;

	if(np->promisc)
	{
		/* Set promiscuous. */
		/* Unconditionally log net taps. */
		DEBUG("Promiscuous mode enabled.\n");
		mc_filter[1] = mc_filter[0] = ~0;
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptAllPhys | AcceptMyPhys;
	}
	else if(np->all_multi)
	{
		mc_filter[1] = mc_filter[0] = ~0;
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
	}
	else
	{
		mc_filter[1] = mc_filter[0] = 0;
		for(i = 0; i < np->nmulti; i++)
		{
			uint32 bit = (ether_crc(6, (unsigned char*)&np->multi[i]) >> 26) & 0x3f;
			if(bit < 32) mc_filter[0] |= (1 << bit);
			else mc_filter[1] |= (1 << (bit - 32));
		}
		rx_mode = AcceptBroadcast | AcceptMulticast | AcceptMyPhys;
	}

	writel(mc_filter[0], ioaddr + MulticastFilter0);
	writel(mc_filter[1], ioaddr + MulticastFilter1);
	np->mcast_filter[0] = mc_filter[0];
	np->mcast_filter[1] = mc_filter[1];

	if((np->txrx_config & RxFilter) != rx_mode)
	{
		np->txrx_config &= ~RxFilter;
		np->txrx_config |= rx_mode;
		writel(np->txrx_config, ioaddr + RxConfig);
	}
}


static void init(mtd80x_private *dev)
{
	uint32 ioaddr = dev->reg_base;
	int i;

	/* Some chips may need to be reset. */

	writel(~0, ioaddr + IntrStatus);

	init_ring(dev);

	writel(dev->rx_ring_addr, ioaddr + RxRingPtr);
	writel(dev->tx_ring_addr, ioaddr + TxRingPtr);

	/* Address register must be written as words. */
#if __INTEL__
	writel(*((uint32*)(&dev->dev_addr.byte[0])), ioaddr + StationAddr);
	writel(*((uint16*)(&dev->dev_addr.byte[4])), ioaddr + StationAddr + 4);
#else
	writel(B_SWAP_INT32(*((uint32*)(&dev->dev_addr.byte[0]))), ioaddr + StationAddr);
	writel(B_SWAP_INT16(*((uint16*)(&dev->dev_addr.byte[4]))), ioaddr + StationAddr + 4);
#endif

	/* Initialize other registers. */
	writel(0x038, ioaddr + PCIBusCfg);

	dev->txrx_config = TxEnable | RxEnable | TxThreshold | 0x0f00;
	dev->mcast_filter[0] = dev->mcast_filter[1] = 0;
	set_rx_mode(dev);

	/* Enable interrupts by setting the interrupt mask. */
	writel(IntrRxDone | IntrRxErr | IntrRxEmpty | IntrTxDone | IntrTxEmpty |
		   StatsMax | RxOverflow | TxUnderrun | IntrPCIErr | NWayDone | LinkChange,
	       ioaddr + IntrEnable);

	dev->last_trans = system_time();

	resume_thread(dev->iothr);
}


/* io_lock gets you exclusive access to the card. */
#define io_lock(data)				(acquire_sem(data->iolock) == B_NO_ERROR)
#define io_unlock(data)				release_sem_etc(data->iolock, 1, 0/*B_DO_NOT_RESCHEDULE*/)

/* output_wait wakes up when the card is ready to transmit another packet */
#define output_wait(data, t)		acquire_sem_etc(data->olock, 1, B_TIMEOUT, t)
#define output_unwait(data, c)		release_sem_etc(data->olock, c, 0/*B_DO_NOT_RESCHEDULE*/)

/* input_wait wakes up when the card has at least one packet on it */
#define input_wait(data)			acquire_sem_etc(data->ilock ,1, B_CAN_INTERRUPT, 0)
#define input_unwait(data, c)		release_sem_etc(data->ilock, c, 0/*B_DO_NOT_RESCHEDULE*/)


static bool my_packet(mtd80x_private *data, unsigned char *addr)
{
	int i;

	if(addr == NULL) return false;
	if(data->promisc || data->all_multi) return true;
	if(memcmp(addr, &data->dev_addr, 6) == 0) return true;
	if(memcmp(addr, "\xff\xff\xff\xff\xff\xff", 6) == 0) return true;
	for(i = 0; i < data->nmulti; i++) if(memcmp(addr, &data->multi[i], 6) == 0) return true;
	return false;
}


/* This routine is logically part of the interrupt handler, but separated for clarity and better register allocation. */
static int netdev_rx(mtd80x_private *np)
{
	uint16 i, pkt_len;
	netdev_desc_rx *desc;
	uint32 desc_status;

	unsigned char *buf;
	netdev_list *list, *last_list;

	for(i = 0, desc = &np->rx_ring[0]; i < RX_RING_SIZE && desc != NULL; i++, desc = &np->rx_ring[i])
	{
		if((desc_status = le32_to_cpu(desc->status)) == DescOwn) continue;

		DEBUG("status was %8.8x.\n", desc_status);

		pkt_len = 0;
		if((desc_status & RxDescWholePkt) != RxDescWholePkt)
		{
			DEBUG("Oversized Ethernet frame spanned multiple buffers, entry %#x length %d status %4.4x!\n",
			      i, desc_status >> 16, desc_status);
		}
		else if(desc_status & RxDescErrSum)
		{
			/* There was a error. */
			DEBUG("Rx error was %8.8x.\n", desc_status);
		}
		else
		{
			/* Reported length should omit the CRC. */
			pkt_len = ((desc_status >> 16) & 0xfff) - 4;
			DEBUG("normal Rx pkt length %d.\n", pkt_len);
		}

		if(!(pkt_len < 6 || desc->rxbuffer == NULL || !my_packet(np, desc->rxbuffer)))
		{
			if(np->rx_list_len >= MAX_QUEUE_LENGTH) break;

			if((buf = (unsigned char*)malloc(pkt_len)) != NULL)
			{
				if((list = (netdev_list*)malloc(sizeof(netdev_list))) != NULL)
				{
					last_list = np->rx_list;
					while(!(last_list == NULL || last_list->next == NULL)) last_list = last_list->next;

					bzero(buf, pkt_len);
					memcpy(buf, desc->rxbuffer, pkt_len);

					list->size = pkt_len;
					list->buf = buf;
					list->next = NULL;
					list->prev = last_list;
					if(last_list != NULL) last_list->next = list;
					else np->rx_list = list;

					np->rx_list_len++;

					input_unwait(np, 1);
				}
				else
				{
					free(buf);
				}
			}
		}

		desc->ctrl_length = cpu_to_le32(np->rx_buf_sz);
		desc->rxbuffer = ((unsigned char*)(np->rx_ring) + RX_RING_SIZE * sizeof(netdev_desc_rx) + np->rx_buf_sz * i);
		desc->buf_addr = cpu_to_le32(np->rx_ring_addr + RX_RING_SIZE * sizeof(netdev_desc_rx) + np->rx_buf_sz * i);
		desc->status = cpu_to_le32(DescOwn);
	}

	/* Restart Rx engine if stopped. */
	writel(0, np->reg_base + RxStartDemand);

	return 0;
}


static int32 mtd80x_interrupt(void *_data)
{
	mtd80x_private *np = (mtd80x_private*)_data;
	/* the handling would do in "mtd80x_thread", we don't need spinlock or others. */
	release_sem_etc(np->iosem, 1, B_DO_NOT_RESCHEDULE);
	return B_UNHANDLED_INTERRUPT;
}


static void check_duplex(mtd80x_private *np)
{
	uint32 ioaddr = np->reg_base;
	int new_tx_mode = np->txrx_config;

	if(np->medialock)
	{
	}
	else
	{
		int mii_reg5 = mdio_read(np, np->phys[0], 5);
		int negotiated = mii_reg5 & np->advertising;
		int duplex = ((negotiated & 0x0100) || (negotiated & 0x01C0) == 0x0040);
		if(np->duplex_lock || mii_reg5 == 0xffff) return;
		if(duplex) new_tx_mode |= TxModeFDX;
		if(np->full_duplex != duplex)
		{
			np->full_duplex = duplex;
			DEBUG("Setting %s-duplex based on MII #%d negotiated capability %4.4x.\n",
				  duplex ? "full" : "half", np->phys[0], negotiated);
		}
	}

	if(np->txrx_config != new_tx_mode) writel(new_tx_mode, ioaddr + RxConfig);
}


static void netdev_error(mtd80x_private *np, int intr_status)
{
	uint32 ioaddr = np->reg_base;

	if(intr_status & (LinkChange | NWayDone))
	{
		DEBUG("Link changed: Autonegotiation advertising %4.4x  partner %4.4x.\n",
			  mdio_read(np, np->phys[0], 4), mdio_read(np, np->phys[0], 5));
		check_duplex(np);
	}

	if(intr_status & TxUnderrun)
	{
		DEBUG("Transmitter underrun.\n");
		writel(0, ioaddr + TxStartDemand);
	}

	if(intr_status & RxOverflow)
	{
		DEBUG("Receiver overflow.\n");
		writel(0, ioaddr + RxStartDemand);
		netdev_rx(np); /* Refill Rx descriptors */
	}

	if(intr_status & ~(LinkChange | NWayDone | StatsMax | TxUnderrun | RxOverflow | TxEarly | RxEarly))
	{
		DEBUG("Something Wicked happened! %4.4x.\n", intr_status);
		DEBUG("tcrrcr=%4.4x, bcr=%4.4x.\n", np->txrx_config, readl(np->reg_base + PCIBusCfg));
	}

	/* Hmmmmm, it's not clear how to recover from PCI faults. */
	if(intr_status & IntrPCIErr)
	{
		const char *const pcierr[4] = {"Parity Error", "Master Abort", "Target Abort", "Unknown Error"};
		DEBUG("PCI Bus %s, %x.\n", pcierr[(intr_status>>11) & 3], intr_status);
	}
}


static void mtd80x_dosomething(void *_data)
{
	mtd80x_private *np = (mtd80x_private*)_data;
	uint32 ioaddr = np->reg_base;
	int i;
	int32 boguscnt = max_interrupt_work;

	if(!io_lock(np)) return;

	while(!(np->interrupted))
	{
		uint32 intr_status = readl(ioaddr + IntrStatus);

		/* Acknowledge all of the current interrupt sources ASAP. */
		writel(intr_status, ioaddr + IntrStatus);

		if(intr_status == 0) break;

		DEBUG("Interrupt, status %4.4x.\n", intr_status);

		if(intr_status & IntrRxEmpty) DEBUG("Receive buffer is empty.\n");

		if(intr_status & IntrRxDone) netdev_rx(np);

		if(np->txfreecount != TX_RING_SIZE)
		{
			np->txfreecount = 0;
			for(i = 0; i < TX_RING_SIZE; i++)
			{
				int tx_status = le32_to_cpu(np->tx_ring[i].status);

				if(tx_status == DescOwn) continue;
				np->txfreecount++;
				if(np->tx_ring[i].buf_addr == 0) continue;

				if(tx_status & (TxErrAbort | TxErrCarrier | TxErrLate | TxErr16Colls | TxErrHeartbeat))
				{
					DEBUG("Transmit error, Tx status %8.8x.\n", tx_status);
				}
				else
				{
					DEBUG("Transmit done, Tx status %8.8x.\n", tx_status);
				}

				np->tx_ring[i].status = 0;
				np->tx_ring[i].ctrl_length = 0;
				np->tx_ring[i].buf_addr = 0;
				output_unwait(np, 1);
			}
		}

		if(intr_status & (IntrRxErr | StatsMax | RxOverflow | TxUnderrun | IntrPCIErr | NWayDone| LinkChange))
			netdev_error(np, intr_status);

		if(--boguscnt < 0)
		{
			DEBUG("Too much work at interrupt, status=0x%4.4x.\n", intr_status);
			break;
		}
	}

	io_unlock(np);
}


static void resetnic(mtd80x_private *np)
{
	int i;
	uint32 ioaddr = np->reg_base;

	/* Stop Tx and Rx */
	i=0;
	while (i<10)
	{
		++i;
		writel(0, ioaddr + RxConfig);
		if(readl(ioaddr + RxConfig) & TxEnable) continue;
		if(readl(ioaddr + RxConfig) & RxEnable) continue;
		break;
	}

	/* reset the nic */
	writel(1, ioaddr + PCIBusCfg);
	do
	{
		if(!(readl(ioaddr + PCIBusCfg) & 0x01)) break;
	}while(1);

	/* Initialize all Rx descriptors. */
	for(i = 0; i < RX_RING_SIZE; i++) np->rx_ring[i].status = cpu_to_le32(DescOwn);

	writel(np->tx_ring_addr, ioaddr + TxRingPtr);
	writel(np->rx_ring_addr, ioaddr + RxRingPtr);

	/* Configure the PCI bus bursts and FIFO thresholds. */
	writel(0x038, ioaddr + PCIBusCfg);

	/* Enable interrupts by setting the interrupt mask. */
	writel(IntrRxDone | IntrRxErr | IntrRxEmpty | IntrTxDone | IntrTxEmpty |
		   StatsMax | RxOverflow | TxUnderrun | IntrPCIErr | NWayDone | LinkChange,
		   ioaddr + IntrEnable);

	/* here we reset the status, otherwise IT'LL ALWAYS be RxOverflow */
	writel(~0, ioaddr + IntrStatus);

	writel(np->txrx_config, ioaddr + RxConfig);
	writel(0, ioaddr + TxStartDemand);
	writel(0, ioaddr + RxStartDemand);
}


static void tx_timeout(mtd80x_private *np)
{
	resetnic(np);
	np->last_trans = system_time();
	return;
}


/* I like thread more than timer, and it also handle something when interrupted */
static int32 mtd80x_thread(void *_data)
{
	mtd80x_private *np = (mtd80x_private*)_data;

	while(true)
	{
		if(acquire_sem_etc(np->iosem, 1, B_TIMEOUT, 100000) == B_OK) mtd80x_dosomething(_data);

		if(!io_lock(np)) continue;
		if(np->interrupted) {io_unlock(np); break;}
		if(np->txfreecount != TX_RING_SIZE && system_time() - np->last_trans > TX_TIMEOUT) tx_timeout(np);
		io_unlock(np);
	}

	return 0;
}


static status_t open_hook(const char *name, uint32 flags, void **cookie)
{
	int32 devID;
	int32 mask;
	status_t status;
	char *devName;
	mtd80x_private *data;

	*cookie = NULL;

	/* Find device name */
	for(devID = 0; (devName = gDevNameList[devID]) != NULL; devID++)
	{
		if(strcmp(name, devName) == 0) break;
	}
	if(devName == NULL) return EINVAL;

	/* Check if the device is busy and set in-use flag if not */
	mask = 1 << devID;
	if(atomic_or(&gOpenMask, mask) & mask) return B_BUSY;

	/* Allocate storage for the cookie */
	if((*cookie = (data = (mtd80x_private *)malloc(sizeof(mtd80x_private)))) == NULL)
	{
		atomic_and(&gOpenMask, ~mask);
		DEBUG("memory allocate failed!\n");
		return B_NO_MEMORY;
	}
	bzero(data, sizeof(mtd80x_private));

	/* Setup the cookie */
	data->devID = devID;
	data->pciInfo = gDevList[devID];

	DEBUG("open %s dev = %x\n", name, data);

	/* enable access to the cards address space */
	if((status = enable_addressing(data)) != B_OK)
	{
		free(data);
		atomic_and(&gOpenMask, ~mask);
		DEBUG("enable_addressing failed!\n");

		*cookie = NULL;
		return status;
	}

	if(probe(data) == 0)
	{
		free(data);
		atomic_and(&gOpenMask, ~mask);
		DEBUG("probe failed!\n");

		*cookie = NULL;
		return B_ERROR;
	}

	if((status = allocate_resources(data)) != B_OK)
	{
#if !__INTEL__
		delete_area(data->ioarea);
#endif
		free(data);
		atomic_and(&gOpenMask, ~mask);
		DEBUG("allocate_resources failed!\n");

		*cookie = NULL;
		return status;
	}

	/* Setup interrupts */
	install_io_interrupt_handler(data->pciInfo->u.h0.interrupt_line, mtd80x_interrupt, *cookie, 0);
	DEBUG("Interrupts installed at %x\n", data->pciInfo->u.h0.interrupt_line);

	/* Init Device */
	init(data);

	return B_NO_ERROR;
}


static status_t close_hook(void *_data)
{
	mtd80x_private *data = (mtd80x_private*)_data;
	status_t status;

	DEBUG("close dev = %x\n", data);

	/* Force pending reads and writes to terminate */
	while(true) {if(io_lock(data)) break;}
	data->interrupted = true;
	input_unwait(data, 1);
	output_unwait(data, 1);
	release_sem_etc(data->iosem, 1, B_DO_NOT_RESCHEDULE);
	io_unlock(data);

	while(data->inrw != 0)
	{
		snooze(1000000);
		DEBUG("still waiting for read/write to finish\n");
	}

	wait_for_thread(data->iothr, &status);

	/* Disable interrupts by clearing the interrupt mask. */
	writel(0x0000, data->reg_base + IntrEnable);

	/* Stop the chip's Tx and Rx processes. */
	writel(0, data->reg_base + RxConfig);

	remove_io_interrupt_handler(data->pciInfo->u.h0.interrupt_line, mtd80x_interrupt, data);
	free_resources(data);

#if !__INTEL__
	delete_area(data->ioarea);
#endif

	atomic_and(&gOpenMask, ~(1 << data->devID));

	return B_NO_ERROR;
}


static status_t free_hook(void *data)
{
	DEBUG("free %s dev = %x\n", data);
	free(data);
	return B_ERROR;
}


static status_t domulti(mtd80x_private *data, char *addr, bool add)
{
	uint32 i, nmulti;
	status_t retVal = B_ERROR;

	if(!io_lock(data)) return B_ERROR;

	if((nmulti = data->nmulti) < MAX_MULTI || !add)
	{
		retVal = B_NO_ERROR;

		for(i = 0; i < nmulti; i++)
		{
			if(memcmp(&(data->multi[i]), addr, 6) == 0) break;
		}

		if(i == nmulti && add)
		{
			memcpy(&(data->multi[i]), addr, 6);
			data->nmulti++;
		}
		else if(i < nmulti && !add)
		{
			if(i != nmulti - 1) memcpy(&data->multi[i], &data->multi[i + 1], (nmulti - i - 1) * 6);
			data->nmulti--;
		}

		if(data->nmulti != nmulti) set_rx_mode(data);
	}

	io_unlock(data);

	return retVal;
}


static status_t control_hook(void *_data, uint32 msg, void *buf, size_t len)
{
	mtd80x_private *data = (mtd80x_private*)_data;
	unsigned int sz;

	if(data == NULL) return B_ERROR;

	switch(msg)
	{
		case ETHER_INIT:
			DEBUG("control: ETHER_INIT\n");
			return B_NO_ERROR;

		case ETHER_GETADDR:
			if(buf == NULL) return B_ERROR;
			DEBUG("control: GET_ADDR\n");
			memcpy(buf, &(data->dev_addr), 6);
			return B_NO_ERROR;

		case ETHER_NONBLOCK:
			if(buf == NULL || !io_lock(data)) return B_ERROR;
			memcpy(&data->nonblocking, buf, sizeof(data->nonblocking));
			DEBUG("control: NON_BLOCK %x\n", data->nonblocking);
			io_unlock(data);
			return B_NO_ERROR;

		case ETHER_ADDMULTI:
		case ETHER_REMMULTI:
			if(buf == NULL) return B_ERROR;
			DEBUG("control: DO_MULTI(%s) - %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n", msg == ETHER_ADDMULTI ? "add" : "remove",
			      *((char*)buf), *((char*)buf + 1), *((char*)buf + 2),
			      *((char*)buf + 3), *((char*)buf + 4), *((char*)buf + 5));
			return(domulti(data, (char*)buf, msg == ETHER_ADDMULTI));

		case ETHER_SETPROMISC:
			if(buf == NULL || !io_lock(data)) return B_ERROR;
			memcpy(&data->promisc, buf, sizeof(data->promisc));
			DEBUG("control: PROMISC %x\n", data->promisc);
			set_rx_mode(data);
			io_unlock(data);
			return B_NO_ERROR;

		case ETHER_GETFRAMESIZE:
			if(buf == NULL) return B_ERROR;
			sz = PKT_BUF_SZ /*(unsigned int)data->rx_buf_sz */;
			DEBUG("control: GET_FRAMESIZE %u\n", sz);
			memcpy(buf, &sz, sizeof(sz));
			return B_NO_ERROR;

		default:
			break;
	}

	return B_ERROR;
}


/* Copy a packet from the queue */
static int copy_packet(mtd80x_private *data, unsigned char *ether_buf, int buflen)
{
	int len = -1;
	netdev_list *list;

	if(ether_buf == NULL || buflen < 0 || !io_lock(data)) return -1;

	while((list = data->rx_list) != NULL)
	{
		if(/*my_packet(data, list->buf) && */list->size > 0)
			if((len = MIN(list->size, buflen)) > 0) memcpy(ether_buf, list->buf, len);

		data->rx_list = list->next;
		if(data->rx_list != NULL) data->rx_list->prev = NULL;

		if(list->buf) free(list->buf);
		free(list);

		data->rx_list_len--;

		if(len >= 0) break;
	}

	if(data->rx_list != NULL) input_unwait(data, 1);

	io_unlock(data);

	return MAX(len, 0);
}


static status_t read_hook(void *_data, off_t pos, void *buf, size_t *len)
{
	mtd80x_private *data = (mtd80x_private*)_data;
	ulong buflen;
	int packet_len = 0;
	bool islocked = false;
	status_t status;

	buflen = *len;
	atomic_add(&data->inrw, 1);

	islocked = io_lock(data);

	if(!islocked || data->interrupted)
	{
		if(islocked) io_unlock(data);

		atomic_add(&data->inrw, -1);
		return B_INTERRUPTED;
	}

	while(islocked && packet_len == 0)
	{
		if(!data->nonblocking)
		{
			io_unlock(data);
			status = input_wait(data);
			islocked = (status == B_OK ? io_lock(data) : false);
		}

		if(!islocked || data->interrupted)
		{
			if(islocked) io_unlock(data);

			atomic_add(&data->inrw, -1);
			return B_INTERRUPTED;
		}

		io_unlock(data);

		packet_len = copy_packet(data, (unsigned char *)buf, buflen);

		if(islocked = io_lock(data))
		{
			if(data->nonblocking) break;
		}
	}

	if(islocked) io_unlock(data);

	atomic_add(&data->inrw, -1);
	*len = packet_len;
	return B_NO_ERROR;
}


static status_t write_hook(void *_data, off_t pos, const void *buf, size_t *len)
{
	mtd80x_private *data = (mtd80x_private*)_data;
	int i;
	status_t status;
	uint32 buflen;
	unsigned char* buffer;
	bool islocked = false;

	buflen = *len;
	atomic_add(&data->inrw, 1);

	islocked = io_lock(data);

	if(!islocked || data->interrupted)
	{
		if(islocked) io_unlock(data);

		atomic_add(&data->inrw, -1);
		return B_INTERRUPTED;
	}

	io_unlock(data);
	/* Wait for somebody else (if any) to finish transmitting */
	status = output_wait(data, ETHER_TRANSMIT_TIMEOUT);
	islocked = io_lock(data);

	if(status < B_NO_ERROR || !islocked || data->interrupted)
	{
		if(islocked) io_unlock(data);

		atomic_add(&data->inrw, -1);
		return(status == B_OK ? B_INTERRUPTED : status);
	}

	if(buflen > PKT_BUF_SZ) buflen = PKT_BUF_SZ;

	for(i = 0; i < TX_RING_SIZE; i++)
	{
		if(data->tx_ring[i].buf_addr != 0) continue;

		buffer = ((unsigned char*)(data->tx_ring) + TX_RING_SIZE * sizeof(netdev_desc_tx) + PKT_BUF_SZ * i);
		memcpy(buffer, buf, buflen);

		data->tx_ring[i].ctrl_length = cpu_to_le32(TxIntrOnDone | TxNormalPkt | (buflen << 11) | buflen);
		data->tx_ring[i].buf_addr = cpu_to_le32(data->tx_ring_addr + TX_RING_SIZE * sizeof(netdev_desc_tx) + PKT_BUF_SZ * i);
		data->tx_ring[i].status = cpu_to_le32(DescOwn);

		/* Wake the potentially-idle transmit channel. */
		writel(0, data->reg_base + TxStartDemand);

		data->txfreecount--;
		data->last_trans = system_time();

		DEBUG("Start transmiting.\n");

		break;
	}

	if(i == TX_RING_SIZE) buflen = 0;

	io_unlock(data);

	atomic_add(&data->inrw, -1);
	*len = buflen;	
	return B_NO_ERROR;
}


/* prototypes */
static device_hooks gDeviceHooks =  {
	open_hook,		/* -> open entry point */
	close_hook,		/* -> close entry point */
	free_hook,		/* -> free entry point */
	control_hook,	/* -> control entry point */
	read_hook,		/* -> read entry point */
	write_hook,		/* -> write entry point */
	NULL,			/* -> select entry point */
	NULL,			/* -> deselect entry point */
	NULL,			/* -> readv */
	NULL			/* -> writev */
};


/* Initalize hardware */
_EXPORT status_t init_hardware(void)
{
	return B_NO_ERROR;
}


/* Initalize driver */
_EXPORT status_t init_driver()
{
	status_t status;
	int32 entries;
	char devName[DEVNAME_LENGTH];
	int32 i;

#if ENABLE_DEBUG
	set_dprintf_enabled(true);
#endif

	if((status = get_module(B_PCI_MODULE_NAME, (module_info **)&gPCIModInfo)) != B_OK)
	{
		DEBUG("get_module failed! %s\n", strerror(status));
		return status;
	}

	/* Find Lan cards */
	if((entries = get_pci_list(gDevList, MAX_CARDS)) == 0)
	{
		DEBUG("cards not found\n");
		free_pci_list(gDevList);
		put_module(B_PCI_MODULE_NAME);
		return B_ERROR;
	}

	/* Create device name list */
	for(i = 0; i < entries; i++)
	{
		sprintf(devName, "%s%ld", kDevDir, i);
		gDevNameList[i] = (char *)malloc(strlen(devName)+1);
		strcpy(gDevNameList[i], devName);
	}
	gDevNameList[i] = NULL;

	return B_OK;
}


/* Uninitalize driver */
_EXPORT void uninit_driver(void)
{
	int32 i;
	void *item;

	/* Free device name list */
	for(i = 0; (item = gDevNameList[i]) != NULL; i++) free(item);

	/* Free device list */
	free_pci_list(gDevList);
	put_module(B_PCI_MODULE_NAME);
}


/* Publish devices */
_EXPORT const char** publish_devices(void)
{
	return((const char **)gDevNameList);
}


/* Device hooks */
_EXPORT device_hooks *find_device(const char *name)
{
	int32 i;
	char *item;

	/* Find device name */
	for(i = 0; (item = gDevNameList[i]) != NULL; i++)
		if(strcmp(name, item) == 0) return(&gDeviceHooks);

	return NULL;
}


_EXPORT int32 api_version = B_CUR_DRIVER_API_VERSION;


#if 0
/* prepare for power event */
static status_t suspend_hook(void *_data)
{
	mtd80x_private *dev = (mtd80x_private*)_data;
	if(!io_lock(dev)) return B_ERROR;
	/* Disable interrupts, stop Tx and Rx. */
	writel(0x0000, dev->reg_base + IntrEnable);
	writel(0, dev->reg_base + RxConfig);
	io_unlock(dev);
	return B_NO_ERROR;
}

/* prepare for power event */
static status_t wake_hook(void *_data)
{
	mtd80x_private *dev = (mtd80x_private*)_data;
	if(!io_lock(dev)) return B_ERROR;
	/* This is incomplete: the actions are very chip specific. */
	set_rx_mode(dev);
	io_unlock(dev);
	return B_NO_ERROR;
}
#endif


_EXPORT void suspend_driver(void)
{
	/* TODO */
}


_EXPORT void wake_driver(void)
{
	/* TODO */
}
